/*
 * windowmanager.cc - Defines the WindowManager class which controls Sapphire's execution.
 * Copyright (C) 2000 Frank Hale
 * frankhale@lycos.com
 * http://sapphire.sourceforge.net/
 *
 * Updated: 6 Nov 2000
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "windowmanager.hh"

WindowManager *wm;

#define SAPPHIRE_KEY_ALT_COUNT 13
KeySym WindowManager::AltKeys[]={XK_0,XK_1,XK_2,XK_3,XK_4,XK_5,XK_6,
				 XK_7,XK_8,XK_9,XK_Page_Up,XK_Page_Down,
				 XK_Tab};

// This function sets up default values for variables, gets command line args,
// sets up our signal handlers, and performs initialization of certain things
// to get ready for the window manager to start managing windows.
WindowManager::WindowManager(int argc, char **argv) 
{
	::wm = this;
	
	font_pix 		= None;
	
	menu_event		= false;
	
	clock_format		= "%m %d %H:%M";
	max_desktops		= 4;
	toolbar_visible		= true;
	toolbar_position	= false;
	
	configurationPath  	= NULL;
	windowmanager_conf 	= NULL;
	display		   	= NULL;

	virtual_desktop_mask	= 1; // starts on first desktop

	XSetWindowAttributes 	attr;
	XColor 			dummy;

	head_client = new Client();
	
	// Allocate memory for our linked lists.
	ClientList    = new LinkedList<Client>;
	menulist      = new LinkedList<BaseMenu>;
	extramenulist = new LinkedList<BaseMenu>;

	for (int i = 1; i < argc; i++) 
	{
		if ((strcmp(argv[i], "-vd")==0) && i+1<argc) {
			max_desktops=atoi(argv[++i]);
			continue;
		}
		if ((strcmp(argv[i], "-display")==0) && i+1<argc)
		{
			display = argv[++i];
			continue;
		}
		if ((strcmp(argv[i], "-version")==0) && i<argc) {
			printf("version %s\nby: %s\nemail: %s\n", windowmanager_version, windowmanager_author, windowmanager_author_email);
			exit(0);
		}
		else {
			printf("usage: %s %s", windowmanager_name, windowmanager_usage);
			exit(1);
		}
	}
        
	// PP: Sun Feb 20 05:36:53 CET 2000
        // If we got the DISPLAY via a cmd line option
        // we should now set it as an evironment variable
        // so that programs we call have it.
        //
        // The other way round also applies, i.e.
        // if there was no cmd line option we should now
        // set the display variable to the value
        // of the environment variable.
	if (display)
        	setenv("DISPLAY", display, 1);
	else
        	display = getenv("DISPLAY");
	     
	XSetErrorHandler(handle_xerror);
	signal(SIGTERM, (void (*)(int)) signalhandler);
	signal(SIGINT,  (void (*)(int)) signalhandler);
	signal(SIGHUP,  (void (*)(int)) signalhandler);

	dpy = XOpenDisplay(display);

	if (!dpy) { 
		err("sorry can't open display %s", display);
		exit(1);
	}

	screen = DefaultScreen(dpy);
	root = RootWindow(dpy, screen);

	depth = DefaultDepth(dpy, screen);
	visual = DefaultVisual(dpy, screen);
		
	XDefineCursor(dpy, root, XCreateFontCursor(wm->dpy, XC_left_ptr));

	// Atoms			
	xa_wm_state        = XInternAtom(dpy, "WM_STATE", False);
	xa_wm_change_state = XInternAtom(dpy, "WM_CHANGE_STATE", False);
	xa_wm_protos       = XInternAtom(dpy, "WM_PROTOCOLS", False);
	xa_wm_delete       = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	xa_wm_cmapwins     = XInternAtom(dpy, "WM_COLORMAP_WINDOWS", False);
	xa_wm_take_focus   = XInternAtom(dpy, "WM_TAKE_FOCUS", False);
	xa_wm_motif_hints  = XInternAtom(dpy, "_MOTIF_WM_HINTS", False);
	
	foreground_color = DEFAULT_FOREGROUND_COLOR;
	background_color = DEFAULT_BACKGROUND_COLOR;
	border_color 	 = DEFAULT_BORDER_COLOR;

	// Colors
	XAllocNamedColor(dpy, DefaultColormap(dpy, screen), foreground_color, &xforeground_color, &dummy);
	XAllocNamedColor(dpy, DefaultColormap(dpy, screen), background_color, &xbackground_color, &dummy);
	XAllocNamedColor(dpy, DefaultColormap(dpy, screen), border_color, &xborder_color, &dummy);

	colors_per_channel = 4;
	image_dither = True;

	image_control = new BImageControl();
  	image_control->installRootColormap(); 

	theme = new Theme();

	getCurrentTheme();

	if(max_desktops > 4 || max_desktops < 1) max_desktops = 4;

	font 	 = theme->getFont();
	menufont = theme->getMenuFont();
	
	xfont = XLoadQueryFont(dpy, font);
	//iheight = 19; // Oops, hardcoded god this sucks
	
	if (!xfont) 
	{ 
		// We should really check here to see if we can 
		// find another font just incase. Dying here without
		// checking again is not cool!
		
		err("the default font was not found, aborting."); 
		exit(-1); 
	}
	
	iheight = xfont->ascent + xfont->descent + 2 * DEFAULT_SPACE;

	font_position = iheight - xfont->descent;

	if(! QueryShapeExtentions())
	{
		printf("What no shape extentions, God you must be using an old X11 release.\n\nHave a nice life\n\nBa Bye...\n\n");	
	}
	
	move_curs   = XCreateFontCursor(dpy, XC_fleur);
	resize_curs = XCreateFontCursor(dpy, XC_plus);
	left_arrow  = XCreateFontCursor(dpy, XC_left_ptr);
	
	createFontPixmap();
	
	gv.font = xfont->fid;
	gv.tile = font_pix;	
	string_gc = XCreateGC(dpy, root,  GCTile | GCFont , &gv);
	XSetFillStyle(dpy, string_gc, FillTiled);

	gv.foreground = xborder_color.pixel;
	gv.line_width = DEFAULT_BORDER_SIZE;
	border_gc = XCreateGC(dpy, root, GCForeground | GCLineWidth, &gv);

	gv.function = GXinvert;
	gv.subwindow_mode = IncludeInferiors;
	invert_gc = XCreateGC(dpy, root, GCFunction | GCSubwindowMode | GCLineWidth, &gv);

	attr.event_mask = SubstructureRedirectMask | 
			  SubstructureNotifyMask   | 
			  ColormapChangeMask       | 
	                  ButtonPressMask          | 
			  ButtonReleaseMask        | 
			  PropertyChangeMask       | 
			  PointerMotionMask        | 
			  FocusChangeMask          |
		          KeyPressMask             |
			  ExposureMask		   ;
			  
	XChangeWindowAttributes(dpy, root, CWEventMask, &attr);
	
	xres = WidthOfScreen(ScreenOfDisplay(wm->dpy, wm->screen));
	yres = HeightOfScreen(ScreenOfDisplay(wm->dpy, wm->screen));
		
	rm = new RootMenu();
	t = new Toolbar();
	im = new IconMenu();
	wmenu = new WindowMenu();

	addToMenuList(rm);
	rm->parseMenuConfigurationFile();

	t->getToolbarMenu()->addMenu("icon menu", im);
	t->getToolbarMenu()->addMenu("window menu", wmenu);
	t->getToolbarMenu()->addMenu("root menu", rm);
	t->getToolbarMenu()->update_all();

	XGrabKey(dpy,XKeysymToKeycode(wm->getDisplay(),XK_Tab),
				      Mod1Mask|ShiftMask,
				      root,True,GrabModeAsync,GrabModeAsync);

	for(int i=0;i<SAPPHIRE_KEY_ALT_COUNT;i++)
		XGrabKey(dpy,XKeysymToKeycode(dpy,AltKeys[i]),Mod1Mask,
			 root,True,GrabModeAsync,GrabModeAsync);

	scan_windows();

	do_event_loop();
	
}

WindowManager::~WindowManager() 
{
	int i;

	image_control->removeImage(font_pix);

	XUngrabKey(dpy,XKeysymToKeycode(dpy,XK_Tab),
		   Mod1Mask|ShiftMask,
		   root);

	for(i=0;i<SAPPHIRE_KEY_ALT_COUNT;i++)
		XUngrabKey(dpy,XKeysymToKeycode(dpy,AltKeys[i]),
			   Mod1Mask,root);

	XFreeFont(dpy, xfont);
	
	XFreeGC(dpy, invert_gc);
	XFreeGC(dpy, string_gc);
	XFreeGC(dpy, border_gc);
	
	for (int i=ClientList->count(); i>0; i--)
	{
		ClientList->remove(i);
	}	
			
	delete t;
	delete rm;
	delete im;

	for (int j=menulist->count(); j>0; j--)
	{
		menulist->remove(j);
	}

	for (int k=extramenulist->count(); k>0; k--)
	{
		extramenulist->remove(k);
	}

	delete menulist;
	delete extramenulist;

	delete theme;

	delete head_client;
	delete ClientList;
	
	delete image_control;
	
	delete [] windowmanager_conf;
	delete [] configurationPath;
}

void WindowManager::getCurrentTheme()
{
	XrmDatabase resource_db;	
	XrmValue value;
	char *value_type;
	FILE* currentTheme;

	windowmanager_conf = new char[strlen(HOME) + 32];
	sprintf(windowmanager_conf, "%s/.sapphire/wmconf", HOME);
	
	if((currentTheme = fopen(windowmanager_conf, "r"))==NULL)
	{
		char* defaultTheme = new char[strlen(configuration_home) + 32];
		sprintf(defaultTheme, "%s/themes/default.theme", configuration_home);

		theme->parseTheme(defaultTheme);
		theme->setToolbarVisible();
		
		delete [] defaultTheme;
				
	} else {
		
		XrmInitialize();

		resource_db = XrmGetFileDatabase(windowmanager_conf);

		if (XrmGetResource(resource_db, "toolbar.visible",
		     "Toolbar.Visible", &value_type, &value))
		{
			if(strcmp(value.addr, "true")==0) 
			{
				theme->setToolbarVisible();
				toolbar_visible=true;
			} else {
				toolbar_visible=false;
			}
		}

		if (XrmGetResource(resource_db, "toolbar.position",
		     "Toolbar.Position", &value_type, &value))
		{
			if(strcmp(value.addr, "top")==0) 
			{
				theme->setToolbarTop();
				toolbar_position = true;
			} else {
				toolbar_position = false;
			}
		}
		
		// Clock format
		if (XrmGetResource(resource_db,"clock.format","Clock.Format",&value_type,&value))
		  {
		    clock_format= new char[value.size];
		    strcpy(clock_format,value.addr);
		  } else {
		    //clock_format="%Y-%m-%d-%H:%M %z";
		    clock_format="%m %d %H:%M";
		  }

		if (max_desktops==16)
			if (XrmGetResource(resource_db,"desktops.count","Desktops.Count",
					   &value_type,&value))
				max_desktops=atoi(value.addr);
	
		if (XrmGetResource(resource_db, "start.theme",
		     "Start.Theme", &value_type, &value))
		{
			char *buffer = new char[value.size+strlen(HOME)];
			
			strcpy(buffer, value.addr);
			
			if (value.addr[0] == '~') 
			{
				sprintf(buffer, "%s%s", HOME, value.addr+1);
			}

			theme->parseTheme(buffer);
			
			delete [] buffer;
		}

		XrmDestroyDatabase(resource_db);
		
		fclose(currentTheme);	
	}
}


void WindowManager::createFontPixmap()
{
	unsigned long style=0;
	
	style = wm->getTheme()->getFontStyle();		

	font_pix = wm->getImageControl()->renderImage(
			"xfont",
			xfont->ascent, 
			xfont->descent, 
			style, 
			wm->getTheme()->getFontColorFrom(), 
			wm->getTheme()->getFontColorTo()
			);
}

// We really should have some way to check if the menu needs
// reconfiguring. It shouldn't do anymore processing than
// it needs to.
void WindowManager::reconfigureMenu()
{
	XGrabServer(dpy);
	
	if(menulist->count())
	{
		BaseMenu* m;
		
		for(int i=menulist->count(); i>=0; i--)
		{
			menulist->remove(i);
		}

		rm->removeAll();		
		addToMenuList(rm);
		rm->parseMenuConfigurationFile();
	}

	XUngrabServer(dpy);
}

int WindowManager::QueryShapeExtentions()
{
	int dummy2;

	return shape = XShapeQueryExtension(dpy, &shape_event, &dummy2);
}

void WindowManager::scan_windows()
{
	unsigned int i, nwins;
	Window dw1, dw2, *wins;
	XWindowAttributes attr;

	XQueryTree(dpy, root, &dw1, &dw2, &wins, &nwins);
	for (i = 0; i < nwins; i++) {
		XGetWindowAttributes(dpy, wins[i], &attr);
		if (!attr.override_redirect && attr.map_state == IsViewable)
			head_client->make_new_client(wins[i]);
	}
	XFree(wins);
}

void WindowManager::handleKeyPress(XEvent *ev) {
	int alt,shift,ctrl;
	KeySym ks;

	alt=(ev->xkey.state&Mod1Mask)?1:0;
	shift=(ev->xkey.state&ShiftMask)?1:0;
	ctrl=(ev->xkey.state&ControlMask)?1:0;

	ks=XKeycodeToKeysym(wm->getDisplay(),ev->xkey.keycode,0);
	if (ks==NoSymbol) return;

	if ((alt)&&(!ctrl)&&(!shift)) {
		switch(ks) {
		case XK_1: goToDesktop(0); break;
		case XK_2: goToDesktop(1); break;
		case XK_3: goToDesktop(2); break;
		case XK_4: goToDesktop(3); break;
		case XK_5: goToDesktop(4); break;
		case XK_6: goToDesktop(5); break;
		case XK_7: goToDesktop(6); break;
		case XK_8: goToDesktop(7); break;
		case XK_9: goToDesktop(8); break;
		case XK_0: goToDesktop(9); break;
		case XK_Page_Up: goToNextDesktop(); break;
		case XK_Page_Down: goToPreviousDesktop();
			break;
		case XK_Tab:
			XCirculateSubwindowsUp(wm->getDisplay(), wm->getRootWindow());
			break;
		}
	}

	if ((alt)&&(!ctrl)&&(shift)) {
		if (ks==XK_Tab)
			XCirculateSubwindowsDown(wm->getDisplay(), wm->getRootWindow());
	}
}

void WindowManager::do_event_loop()
{
	XEvent ev;

	BaseMenu *m = None;

	int xfd = ConnectionNumber(wm->dpy);

	int dx=0,dy=0;
	int alt,shift,ctrl;

	for (;;) {

		if(XPending(wm->dpy))
		{
			XNextEvent(wm->dpy, &ev);

			switch (ev.type) {
			        case KeyPress:
					handleKeyPress(&ev);
					break;
			        case ButtonPress:
				{
					head_client->handle_button_press_event(&ev.xbutton); 
					t->do_button_press_event(&ev.xbutton);
					
					if(m = findInMenuList(&ev.xbutton.window))
					{
						m->handle_button_press_event(&ev.xbutton);	
					}
				
					switch(ev.xbutton.button)
					{
						case 2: {
						
					 		if(ev.xbutton.window == root)
					  		{
								t->setStartMenuButtonStateNotPressed();
								// was im instead of wmenu
								wmenu->update_all();
								wmenu->setAllMenuTitlesVisible(wmenu,True);
								wmenu->show();
					  		}
							
						break;
						}

						case 3: {
					 		if(ev.xbutton.window == root)
					  		{
								t->setStartMenuButtonStateNotPressed();
								rm->hideAllVisibleSubmenus();
								rm->setAllMenuTitlesVisible(rm, True);
								rm->show();
					  		}
						break;	
						}
						
						default:
						{
							/* this if (from the if to the closing
                                                           brace) was the source of all trouble
                                                           merging patches 
                                                           -- fixed by hand (bergo)*/
							if(ev.xbutton.window == root)
					  		{
								t->setStartMenuButtonStateNotPressed();
								// was im instead of wmenu
								wmenu->setAllMenuTitlesVisible(wmenu, False);
								wmenu->hide(wmenu);
								rm->setAllMenuTitlesVisible(rm, False);
								rm->hideAllVisibleSubmenus();
								rm->hide(rm);
							}

						}
						break;

					}
				break;
				}
			
				case ButtonRelease:
				{
					head_client->handle_button_release_event(&ev.xbutton); 
					
					if(m = findInMenuList(&ev.xbutton.window))
					{
						m->handle_button_release_event(&ev.xbutton);	
					}
										
					switch(ev.xbutton.button)
					{
					        case 1:
						case 2: 
						case 3: 
						default:
						{
							if(menu_event)
							{
								wm->ungrabPointer();
								menu_event=false;
							}
						
						}
						break;
					}
					
					
				break;
				}
				
				case ConfigureRequest:
				{
					head_client->handle_configure_request(&ev.xconfigurerequest); 
				
				break;
				}
				
				case ConfigureNotify:
				{
				break;
				}
				
				case DestroyNotify:
				{
					head_client->handle_destroy_notify(&ev.xdestroywindow);
				break;
				}
				
				case FocusIn:
				{
				
					head_client->handle_focus_in_event(&ev.xfocus);
				break;
				}
				
				case FocusOut:
				{
				
					head_client->handle_focus_out_event(&ev.xfocus);

				break;
				}
				
				case MapRequest:
				{
					head_client->handle_map_request(&ev.xmaprequest);
				
				break;
				}
			
				case UnmapNotify:
				{
					head_client->handle_unmap_event(&ev.xunmap); 
				break;
				}
			
				case ClientMessage:
				{
					head_client->handle_client_message(&ev.xclient); 
				break;
				}
			
				case ColormapNotify:
				{
					head_client->handle_colormap_change(&ev.xcolormap); 
				break;
				}
			
				case PropertyNotify:
				{
					head_client->handle_property_change(&ev.xproperty); 
				break;
				}

				case LeaveNotify:
				{									
					if(m = findInMenuList(&ev.xcrossing.window))
					{
						m->handle_leave_notify(&ev.xcrossing);	
					}

				break;
				}
			
				case EnterNotify:
				{
					head_client->handle_enter_event(&ev.xcrossing); 
										
					if(m = findInMenuList(&ev.xcrossing.window))
					{
						m->handle_enter_notify(&ev.xcrossing);	
					}
				break;
				}

				case MotionNotify:
				{
					if(menu_event)
					{
						
						grabPointer(root, ButtonReleaseMask | PointerMotionMask, move_curs);
						dx = ev.xmotion.x_root - menu->menu.x_move;
						dy = ev.xmotion.y_root - menu->menu.y_move;
						menu->menu.x = dx;
						menu->menu.y = dy;
						XMoveWindow(dpy, menu->menu.window, dx, dy);
										
					}
					
					if(m = findInMenuList(&ev.xmotion.window))
					{
						m->handle_motion_notify_event(&ev.xmotion);
					}
					
					head_client->handle_motion_event(&ev.xmotion); 

				break;
				}
						
				case Expose:
				{
					head_client->handle_expose_event(&ev.xexpose);
					t->handle_expose_event(&ev.xexpose); 
				
					rm->handle_expose_event(&ev.xexpose); 
					
					if(m = findInMenuList(&ev.xexpose.window))
					{
						m->handle_expose_event(&ev.xexpose); 
					}
				
					break;
				}
											
				case ReparentNotify:
				{
					head_client->handle_reparent_event(&ev.xreparent); 
				
				break;
				}
				
				default: 
				{				
					if (wm->shape && ev.type == wm->shape_event) 
					head_client->handle_shape_change((XShapeEvent *)&ev);
				break;		 
				}	
			
			
			}
		
		} else {

			fd_set rfds;
			FD_ZERO(&rfds);
			FD_SET(xfd, &rfds);

			struct timeval tv;
			tv.tv_sec = 0;
			tv.tv_usec = 1000;
			
			select(xfd + 1, &rfds, 0, 0, &tv);

			t->checkClock();
		}

	}
	

}

// Function:
//
// void WindowManager::quit_nicely()
//
// Purpose:
//
// This functions sole purpose in life is to preform clean up so that we can
// shutdown Sapphire safely.
//
// Returns:
// 
// This function returns nothing.
//
void WindowManager::quit_nicely()
{
	saveCurrentTheme();
	
	this->~WindowManager();

	XSetInputFocus(dpy, PointerRoot, RevertToPointerRoot, CurrentTime);
	XInstallColormap(dpy, DefaultColormap(dpy, screen));
	XCloseDisplay(dpy);
	
	printf("%s", windowmanager_exit_message);
	
	exit(0);
}

void WindowManager::restart()
{
	saveCurrentTheme();
	
	this->~WindowManager();
	
	XSetInputFocus(dpy, PointerRoot, RevertToPointerRoot, CurrentTime);

	XInstallColormap(dpy, DefaultColormap(dpy, screen));

	XCloseDisplay(dpy);
		
	execl("/bin/sh", "sh", "-c", windowmanager_name, 0);
}

BaseMenu* WindowManager::findInMenuList(Window *w)
{
	if(extramenulist->count())
	{
		LinkedListIterator<BaseMenu> mx(extramenulist);
				
		for(; mx.current(); mx++) 
		{
			if(*w == mx.current()->getMenuWindow()      || 
			   *w == mx.current()->getMenuTitleWindow() ||
			   *w == mx.current()->getMenuItemWindow())
			{
				return mx.current();
			}
		}	
	
	} 
	
	if(menulist->count())
	{
		LinkedListIterator<BaseMenu> my(menulist);
				
		for(; my.current(); my++) 
		{
			if(*w == my.current()->getMenuWindow()      || 
			   *w == my.current()->getMenuTitleWindow() ||
			   *w == my.current()->getMenuItemWindow())
			{
				return my.current();
			}
		}
	}
	
	return NULL;
}

void WindowManager::grabPointer(Window w, unsigned int mask, Cursor curs)
{
	if (! XGrabPointer(dpy, w, False, mask, GrabModeAsync, GrabModeAsync, None, curs, CurrentTime) == GrabSuccess ) 
		return;	
}

void WindowManager::ungrabPointer()
{
	XUngrabPointer(dpy, CurrentTime);	
}

void WindowManager::changeTheme(char* themeFile)
{	
	image_control->removeAllImages();	
	
	theme->parseTheme(themeFile);

	XFreeFont(dpy, xfont);
	XFreeGC(dpy, string_gc);
	if(font_pix) XFreePixmap(dpy, font_pix);
	
	xfont = XLoadQueryFont(dpy, theme->getFont());
	
	if(!xfont) 
	{ 
		err("the font specified in the current theme was not found, using default font.");
		
		xfont = XLoadQueryFont(dpy, font);
	}
	
	font_position = iheight - xfont->descent;
	
	createFontPixmap();
	
	gv.font = xfont->fid;
	gv.tile = font_pix;	
	string_gc = XCreateGC(dpy, root,  GCTile | GCFont , &gv);
	XSetFillStyle(dpy, string_gc, FillTiled);
	
	t->changeTheme();	
	
	if(ClientList->count())
	{
		LinkedListIterator<Client> c(ClientList);
		for(; c.current(); c++) 
		{
			c.current()->themeChange(c.current());
		}
	}
	
	if(menulist->count())
	{
		LinkedListIterator<BaseMenu> mx(menulist);
		for(; mx.current(); mx++) 
		{
			mx.current()->reconfigure();
		}
	}

	if(extramenulist->count())
	{
		LinkedListIterator<BaseMenu> my(extramenulist);
		for(; my.current(); my++) 
		{
			my.current()->reconfigure();
		}
	}
}

void WindowManager::saveCurrentTheme()
{
	// Hopefully I've gotten the bugs out of this function
	// Only time will tell....

	char* save_theme = theme->getCurrentTheme();
	
	FILE *wmconf;
	
	if((wmconf = fopen(windowmanager_conf, "w"))==NULL)
	{
		mkdir("/home/frank/.sapphire", S_IRWXU);

		if((wmconf = fopen(windowmanager_conf, "w"))==NULL)
		{
			perror("error:\n");
			exit(-1);
		} 
	}
	
	if(toolbar_visible)
		fprintf(wmconf, "toolbar.visible: true\n");
	else 
		fprintf(wmconf, "toolbar.visible: false\n");
	
	if(toolbar_position)
		fprintf(wmconf, "toolbar.position: top\n");
	else 
		fprintf(wmconf, "toolbar.position: bottom\n");

	fprintf(wmconf, "start.theme: %s\n", save_theme);		
	fprintf(wmconf, "clock.format: %s\n", clock_format);
	fprintf(wmconf, "desktops.count: %d\n", max_desktops);

	fclose(wmconf);
}

int WindowManager::getCurrentDesktopMask() {
	return(virtual_desktop_mask);
}

void WindowManager::goToDesktop(int desktop) {
	if (desktop>=max_desktops) return;
	changeDesktop(1<<desktop);
	t->desktopChanged();
}

void WindowManager::goToNextDesktop() {
	int new_mask;
	new_mask=virtual_desktop_mask<<1;
	if (new_mask == (1<<max_desktops) )
		new_mask=1;
	changeDesktop(new_mask);
	t->desktopChanged();
}

void WindowManager::goToPreviousDesktop() {
	int new_mask;
	new_mask=virtual_desktop_mask>>1;
	if (new_mask == 0 )
		new_mask=1<<(max_desktops-1);
	changeDesktop(new_mask);
	t->desktopChanged();
}

void WindowManager::changeDesktop(int new_mask) {
	int i;
	char r;

	if (new_mask==virtual_desktop_mask) return;

	virtual_desktop_mask=new_mask;	

	if(ClientList->count())
	{
		LinkedListIterator<Client> c(ClientList);
		for(; c.current(); c++) 
			c.current()->updateVisibility(virtual_desktop_mask);
	}
	r=0;
	i=virtual_desktop_mask;
	while((i&0x01)==0) {
	  ++r;
	  i>>=1;
	}
}

void WindowManager::insertWindowMenuEntry(Client *c,char *name,WindowState _state) {
  if (name)
    wmenu->addWindow(name,c,_state);
  else
    wmenu->addWindow(" ",c,_state);
}

void WindowManager::removeWindowMenuEntry(Client *c) {
	wmenu->removeWindow(c);
}
